{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Autor       : Bert Paredis
Module      : HttpReader.pas
Description : This component read files from Internet, using HTTP protocol
Creation    : 12.05.98
Version     : 1.0
Mail        : louis.jansen@skynet.be
Thanks To   : Pasha A. Poskraykoff (pap@savva.ru)

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit HttpReader;

interface

uses
  Windows, Messages, SysUtils, Classes, WinInet, Forms;

type
  Char1024 = array [0..1024] of Char;

  TReadProgress = procedure(Sender: TObject; Const TotalSize, Readed: Integer;
    Const FileName: String) of object;
  TReadFinish   = procedure(Sender: TObject; Const FileSize: Integer;
    Const FileName: String) of object;

  THttpReaderThread = class(TThread)
  private
    FTURL : string;
    FTPath: string;
    FTRPrg: TReadProgress;
    FTRslt: Boolean;
    FTFN  : string ;
    FTFS  : Integer;
  protected
    procedure Execute; override;
  public
    constructor Create;
  end;

  TCustomHttpReader = class(TComponent)
  private
    FURL : string;
    FPath: string;
    FRPrg: TReadProgress;
    FDone: TReadFinish  ;
    FThrd: THttpReaderThread;
    FErr : TNotifyEvent;
    FRslt: Boolean;
    procedure SetPath(Const sPath: String);
    procedure ThreadDone(Sender: TObject);
  public
    function Execute: Boolean;

    property URL       : String read FURL  write FURL   ;
    property LocalPath : String read FPath write SetPath;

    property OnProgress: TReadProgress read FRPrg write FRPrg;
    property OnCopired : TReadFinish   read FDone write FDone;
    property OnError   : TNotifyEvent  read FErr  write FErr ;
  end;

  THttpReader = class(TCustomHttpReader)
  published
    property URL       ;
    property LocalPath ;
    property OnProgress;
    property OnCopired ;
    property OnError   ;
  end;

procedure Register;

implementation

//  TCustomHttpReader

procedure TCustomHttpReader.SetPath(Const sPath: String);
begin
  if FPath <> sPath then begin
    FPath := sPath;
    if Length(FPath) > 0 then
      if FPath[Length(FPath)] <> '\' then
        FPath := FPath + '\';
  end;
end;

function TCustomHttpReader.Execute: Boolean;
begin
  Result := False;
  if Not Assigned(FThrd) then begin
    FThrd := THttpReaderThread.Create;
    with FThrd do begin
      FTURL       := FURL ;
      FTPath      := FPath;
      FTRPrg      := FRPrg;
      OnTerminate := ThreadDone;
    end;
    while Assigned(FThrd) do Application.ProcessMessages;
    Result := FRslt;
  end;
end;

procedure TCustomHttpReader.ThreadDone(Sender: TObject);
begin
  FRslt := FThrd.FTRslt;
  if FRslt then begin
    if Assigned(FDone) then
      FDone(Self, FThrd.FTFS, FThrd.FTFN);
    end
  else begin
    if Assigned(FErr) then
      FErr(Self);
  end;
  FThrd := nil;
end;

//  THttpReaderThread

constructor THttpReaderThread.Create;
begin
  FreeOnTerminate := True;
  inherited Create(False);
end;

procedure THttpReaderThread.Execute;
  procedure ExtractUrlFile(Const eufURL: String;
    Var eufServer, eufFile: String);
    procedure ReplaceChar(Const Char1, Char2: Char; Var _Str: String);
    Var
      I: Integer;
    begin
      for I := 1 to Length(_Str) do
        if _Str[I] = Char1 then _Str[I] := Char2;
    end;
  Var
    S: String ;
    I: Integer;
  begin
    S := eufURL;
    ReplaceChar('/', '\', S);

    eufFile   := ExtractFileName(S);
    eufServer := ExtractFilePath(S);
    ReplaceChar('\', '/', eufFile  );
    ReplaceChar('\', '/', eufServer);

    if (Length(eufServer) > 0) And (eufServer[Length(eufServer)] = '/') then
      SetLength(eufServer, Length(eufServer) -1);

    I := Pos('HTTP://', UpperCase(eufServer));
    if I <> 0 then System.Delete(eufServer, 1, 7);
  end;
var
  hSession : HInternet;
  hConnect : HInternet;
  hRequest : HInternet;
  sServer  : String   ;
  sFile    : String   ;
  fFIle    : TextFile ;
  lpBuf    : Pointer  ;
  dwBufLen : DWORD    ;
  dwIndex  : DWORD    ;
  lpData   : Char1024 ;
  dwBtRead : DWORD    ;
  I        : Integer  ;
  dwReaded : DWORD    ;
begin

  ExtractUrlFile(FTUrl, sServer, sFile);

  hSession := InternetOpen('InternetReader',
    INTERNET_OPEN_TYPE_PRECONFIG, nil, nil, 0);

  hConnect := InternetConnect(hSession, PChar(sServer),
    INTERNET_DEFAULT_HTTP_PORT, nil, nil, INTERNET_SERVICE_HTTP, 0, 0);

  hRequest := HttpOpenRequest(hConnect, 'GET', PChar(sFile), 'HTTP/1.0', nil,
    nil, INTERNET_FLAG_RELOAD, 0);

  HttpSendRequest(hRequest, nil, 0, nil, 0);

  dwIndex  := 0;
  dwBufLen := 1024;
  GetMem(lpBuf, dwBufLen);
  FTRslt := HttpQueryInfo(hRequest, HTTP_QUERY_CONTENT_LENGTH, lpBuf, dwBufLen,
    dwIndex);

  if FTRslt then begin

    FTFS     := StrToInt(StrPas(lpBuf));
    dwReaded := 0;

    FTFN  := FTPath +sFile;

    AssignFile(fFIle, FTFN);
    Rewrite   (fFIle);

    while True do begin

      if Not InternetReadFile(hRequest, @lpData, SizeOf(lpData),
        dwBtRead) then Break
      else begin

        if dwBtRead = 0 then Break
        else begin

          for I := 0 to dwBtRead -1 do Write(fFIle, lpData[I]);
          dwReaded := dwReaded +dwBtRead;

          if Assigned(FTRPrg) then FTRPrg(Self, FTFS, dwReaded, FTFN);

        end;

      end;

    end;

    CloseFile(fFIle);
                   
  end;

  FreeMem(lpBuf);

  InternetCloseHandle(hRequest);
  InternetCloseHandle(hConnect);
  InternetCloseHandle(hSession);

end;

// Register

procedure Register;
begin
  RegisterComponents('PAP', [THttpReader]);
end;

end.
